#include "runtime/renderer_api.h"
#include "runtime/config.h"
#include "runtime/handles.h"
#include "runtime/threading.h"

#include "gpu.h"
#include "pipelines.h"

#include <volk/volk.h>
#include <stdlib.h>

VY_CVAR_I(r_VkMaxPipelineCount, "Maximum number of pipeline objects. Default: 1024", 1024);

typedef struct vy_pipeline_s {
    uint32_t version;
    vy_pipeline pipeline;
    struct vy_pipeline_s *next_free;
} vy_pipeline_slot;

static vy_pipeline_slot *_pipelines;
static vy_pipeline_slot *_first_free;
static vy_rwlock _lock;

vy_result InitPipelineManagement(void) {
    vy_create_rwlock_result lock_res = vyCreateRWLock();
    if (!lock_res.ok)
        return VY_UNKNOWN_ERROR;
    _lock = lock_res.lock;

    _pipelines = calloc(r_VkMaxPipelineCount.i, sizeof(vy_pipeline_slot));
    if (!_pipelines) {
        vyDestroyRWLock(&_lock);
        return VY_OUT_OF_MEMORY;
    }
    /* Keep [0] unused to preserve 0 as the invalid handle */
    _first_free = &_pipelines[1];
    for (int i = 1; i < r_VkMaxPipelineCount.i - 1; ++i) {
        _pipelines[i].next_free = &_pipelines[i + 1];
    }
    return VY_SUCCESS;
}

static void DestroyPipeline(vy_pipeline_slot *slot) {
    if (slot->pipeline.pipeline) {
        vkDestroyPipeline(g_gpu.device, slot->pipeline.pipeline, g_gpu.alloc_cb);
    }
    slot->next_free = _first_free;
    _first_free     = slot;
}

void ShutdownPipelineManagement(void) {
    for (int i = 1; i < r_VkMaxPipelineCount.i; ++i) {
        DestroyPipeline(&_pipelines[i]);
    }
    free(_pipelines);
    vyDestroyRWLock(&_lock);
    _first_free = NULL;
}

vy_pipeline_handle VY_RENDERER_API_FN(CompilePipeline)(const vy_pipeline_info *info) {
    vy_pipeline_handle handle = VY_INVALID_HANDLE;
    vyLockWrite(&_lock);
    if (!_first_free) {
        vyLog("VK", "No free pipeline slots!");
        vyUnlockWrite(&_lock);
        return handle;
    }
    vy_pipeline_slot *slot = _first_free;
    _first_free            = slot->next_free;
    slot->version = (slot->version + 1) & VY_GFX_HANDLE_MAX_VERSION;
    vyUnlockWrite(&_lock);

    /* No other thread that calls compile gets the same slot.
     * Another thread accessing the slot via GetPipeline would get a version mismatch.
     * The same holds for DestroyPipeline
     */

    return handle;
}

void VY_RENDERER_API_FN(DestroyPipeline)(vy_pipeline_handle handle) {
    if (handle.index >= (uint32_t)r_VkMaxPipelineCount.i)
        return;
    vyLockWrite(&_lock);
    if (_pipelines[handle.index].version == handle.version)
        DestroyPipeline(&_pipelines[handle.index]);
    else
        vyLog("VK", "Tried to destroy a pipeline using an outdated handle.");
    vyUnlockWrite(&_lock);
}

const vy_pipeline *vyGetPipeline(vy_pipeline_handle handle) {
    if (handle.index >= (uint32_t)r_VkMaxPipelineCount.i)
        return NULL;
    vyLockRead(&_lock);
    vy_pipeline *res = NULL;
    if (_pipelines[handle.index].version == handle.version)
        res = &_pipelines[handle.index].pipeline;
    else
        vyLog("VK", "Tried to access a pipeline using an outdated handle.");
    vyUnlockRead(&_lock);
    return res;
}
